#include "taskmanager.h"

// TASK THREAD
TaskThread::~TaskThread()
{
    waitUntilOver();
    for(std::thread * t : m_threads)
        delete t;
}

void TaskThread::addTask(Task* task)
{
    m_tasks.push_back(task);
    ++m_totalTasks;
}

void TaskThread::start(int nbThreads)
{
    for(int i=0; i<nbThreads; ++i)
    {
        std::thread *thread = new std::thread(&TaskThread::run, this, i);
        m_currentTasks.push_back(nullptr);
        m_threads.push_back(thread);
    }
}

void TaskThread::run(int id)
{
    m_mutex.lock();
    while(!m_tasks.empty())
    {
        m_currentTasks[id] = m_tasks.front();
        m_tasks.pop_front();

        m_mutex.unlock();
        m_currentTasks[id]->performTask();
        m_mutex.lock();

        delete m_currentTasks[id];
        m_currentTasks[id] = nullptr;
        ++m_tasksPerformed;
    }
    m_mutex.unlock();
}

void TaskThread::waitUntilOver()
{
    for(std::thread * t : m_threads)
        t->join();
}

bool TaskThread::isOver()
{
    return m_tasksPerformed == m_totalTasks;
}

float TaskThread::getProgressProportion()
{
    float progress = 1.f;
    m_mutex.lock();
    if(!isOver())
    {
        progress = m_tasksPerformed;
        for(Task* t : m_currentTasks)
        {
            if(t != nullptr)
                progress += t->getProgressProportion();
        }
        progress /= m_totalTasks;
    }
    m_mutex.unlock();
    return progress;
}

// TASK MANAGER
void TaskManager::addTask(Task* task)
{
    int flags = task->getFlags();
    if(flags & Task::GPU)
        m_gpuTasks.addTask(task);
    else if(flags & Task::HDD)
        m_hddTasks.addTask(task);
    else if(flags & Task::NET)
        m_netTasks.addTask(task);
    else
        m_cpuTasks.addTask(task);
}

void TaskManager::start()
{
    m_hddTasks.start(1); // Reading multiple files simultaneously makes no sense
    m_cpuTasks.start(4);
    m_gpuTasks.start(1); // GPU tasks must be regouped on only one thread to easily have access to an OpenGL context
    m_netTasks.start(4);
}

void TaskManager::waitUntilOver()
{
    m_hddTasks.waitUntilOver();
    m_cpuTasks.waitUntilOver();
    m_gpuTasks.waitUntilOver();
    m_netTasks.waitUntilOver();
}

bool TaskManager::isOver()
{
    return m_hddTasks.isOver()
        && m_cpuTasks.isOver()
        && m_gpuTasks.isOver()
        && m_netTasks.isOver();
}

float TaskManager::getProgress(int flags)
{
    float progress = 0;
    int nbTypes = 0;
    if(flags & Task::GPU)
        progress += m_gpuTasks.getProgressProportion(), ++nbTypes;
    else if(flags & Task::HDD)
        progress += m_hddTasks.getProgressProportion(), ++nbTypes;
    else if(flags & Task::NET)
        progress += m_netTasks.getProgressProportion(), ++nbTypes;
    else
        progress += m_cpuTasks.getProgressProportion(), ++nbTypes;
    if(nbTypes != 0)
        return progress/nbTypes;
    else
        return 1.f;
}
